from typing import Union, List, Tuple
import numpy as np
import pythia as pt


class Gaussian:
    """Gaussian likelihood function for differentiable forward problem.

    Assemble the Gaussian likelihood

    .. math::
        \\mathcal{L}(x) = \\frac{1}{(2\\pi)^{M/2}\\sqrt{\\det \\Sigma}} \\exp\\Bigl( -\\frac{1}{2} \\Vert \\Sigma^{-1/2}(f(x)-\\delta) \\Vert \\Bigr)

    for covariance matrix
    :math:`\\Sigma = \\operatorname{diag}(\\sigma_1(y_1),\\dots,\\sigma_M(y_M))`
    and measurement/observation :math:`\\delta`.

    Parameters
    ----------
    f : function
        Forward model.
    sigma : function
        Error model for standard deviation describing :math:`\\Sigma`.
    xdim : int
        Number of stochastic parameters.
    """

    def __init__(self, f, sigma, xdim):
        """ Initialize Gaussian likelihood object. """
        self._f = f
        self._std = sigma
        self._xdim = xdim

    def likelihood(self, x, y_meas):
        """ Evaluate the Gaussian likelihood with specified measurement.

        Parameters
        ----------
        x : array_like
            Realizations of stochastic parameters.
        y_meas : array_like
            Measurement :math:`\\delta`.
        """
        # check dimensionality and reshape if necessary.
        y, f_val, std, mdim, fdim = self._reshape(x, y_meas)

        a = (2*np.pi)**(-mdim*fdim/2)  # float
        b = 1/np.prod(std, axis=(0, 2))**mdim  # shape: (#points,)
        c = np.prod(np.exp(-(f_val-y)**2/(2*std**2)),
                    axis=(0, 2))  # (#points,)

        return a*b*c

    def log_likelihood(self, x, y_meas):
        """ Evaluate the Gaussian log-likelihood with specified measurement.

        Parameters
        ----------
        x : array_like
            Realizations of stochastic parameters.
        y_meas : array_like
            Measurement :math:`\\delta`.
        """
        # check dimensionality and reshape if necessary.
        y, f_val, std, mdim, fdim = self._reshape(x, y_meas)

        a = -mdim*fdim/2 * np.log(2*np.pi)  # float
        b = -mdim * np.sum(np.log(std), axis=(0, 2))  # (#points,)
        c = np.sum(-(f_val-y)**2/(2*std**2), axis=(0, 2))  # (#points,)

        return a+b+c

    def _reshape(self, x, y_meas):
        """ Check shape compatibility of input and reshape if necessary.

        Parameters
        ----------
        x : array_like
            Realizations of stochastic parameters.
        y_meas : array_like
            Measurement :math:`\\delta`.
        """
        x_val = np.array(x)
        y_val = np.array(y_meas)
        # x shape: (#points, xdim)
        if x_val.ndim < 2:
            x_val.shape = 1, -1
        assert x_val.shape[1] == self._xdim
        # y_meas shape: (mdim, fdim)
        if y_val.ndim < 2:
            y_val.shape = 1, -1
        assert x_val.ndim == 2 and y_val.ndim == 2

        # get number of measurements and image dim of f
        mdim, fdim = y_val.shape

        f_val = self._f(x_val)
        # f_val shape: (#points, fdim)
        if f_val.ndim < 2:
            f_val.shape = 1, -1
        assert f_val.shape[1] == fdim

        std = self._std(x_val)
        # std shape: (#points, fdim)
        if std.ndim < 2:
            std.shape = 1, -1
        if std.shape[1] == 1:
            std = std*np.ones(fdim)
        assert std.shape[1] == fdim
        assert np.min(std) >= 0

        # Reshape for fast multiplication
        y = np.expand_dims(y_meas, axis=1)  # (mdim, 1, fdim)
        std = np.expand_dims(self._std(x_val), axis=0)  # (1, #points, fdim)
        f_val = np.expand_dims(f_val, axis=0)  # (1, #points, fdim)

        return y, f_val, std, mdim, fdim


class Batman:
    """Target function for tutorial 1."""

    def __init__(self) -> None:
        """Target function for tutorial 1."""

    def _f(self, y: np.ndarray) -> np.ndarray:
        """Helper function."""
        val = (np.abs(y/2)
               - (3*np.sqrt(33)-7)/122 * y**2
               - 3
               + np.sqrt(1-(np.abs(np.abs(y)-2)-1)**2)
               )
        return val

    def _g(self, y: np.ndarray) -> np.ndarray:
        """Helper function."""
        a = -(np.abs(y)-1)*(np.abs(y)-0.75)
        b = np.zeros_like(a)
        idx = np.where(np.abs(a) <= 1e-14)[0]
        b[idx] = 1
        idx = np.where(np.abs(a) >= 1e-14)[0]
        b[idx] = np.abs(a[idx])/a[idx]
        val = 9*np.sqrt(b) - 8*np.abs(y)
        return val

    def _h(self, y: np.ndarray) -> np.ndarray:
        """Helper function."""
        a = -(np.abs(y)-0.5)*(np.abs(y)-0.75)
        val = 3*np.abs(y) + 0.75*np.sqrt(np.abs(a)/a)
        return val

    def _p(self, y: np.ndarray) -> np.ndarray:
        """Helper function."""
        a = -(y-0.5)*(y+0.5)
        val = 2.25*np.sqrt(np.abs(a)/a)
        return val

    def _q(self, y: np.ndarray) -> np.ndarray:
        """Helper function."""
        a = 6*np.sqrt(10)/7
        b = (1.5 - 0.5*np.abs(y)) * np.sqrt(np.abs(np.abs(y)-1)/(np.abs(y)-1))
        c = 6*np.sqrt(10)/14 * np.sqrt(4-(np.abs(y)-1)**2)
        return a + b - c

    def _r(self, y: np.ndarray) -> np.ndarray:
        """Helper function."""
        a = np.sqrt(1-(y/7)**2)
        b = np.sqrt(np.abs(np.abs(y)-3)/(np.abs(y)-3))
        return 3 * a * b

    def _s(self, y: np.ndarray) -> np.ndarray:
        """Helper function."""
        a = np.sqrt(1-(y/7)**2)
        b = np.sqrt(np.abs(np.abs(y)-4)/(np.abs(y)-4))
        return -3 * a * b

    def _t(self, y: np.ndarray) -> np.ndarray:
        """Helper function."""
        return np.sqrt(np.abs(y)/y)

    def upper(self, y: np.ndarray) -> np.ndarray:
        """Upper part of the batman function."""
        val = np.zeros(y.size)

        # set r parts
        idx = np.where(np.logical_and(y < -3, y >= -7))
        val[idx] = self._r(y[idx])
        idx = np.where(np.logical_and(y <= 7, y > 3))
        val[idx] = self._r(y[idx])

        # set q parts
        idx = np.where(np.logical_and(y < -1, y >= -3))
        val[idx] = self._q(y[idx])
        idx = np.where(np.logical_and(y <= 3, y > 1))
        val[idx] = self._q(y[idx])

        # set g parts
        idx = np.where(np.logical_and(y <= -0.75, y >= -1))
        val[idx] = self._g(y[idx])
        idx = np.where(np.logical_and(y <= 1, y >= 0.75))
        val[idx] = self._g(y[idx])

        # set h parts
        idx = np.where(np.logical_and(y < -0.5, y > -0.75))
        val[idx] = self._h(y[idx])
        idx = np.where(np.logical_and(y < 0.75, y > 0.5))
        val[idx] = self._h(y[idx])

        # set p parts
        idx = np.where(np.logical_and(y >= -0.5, y <= 0.5))
        val[idx] = self._p(y[idx])

        return val + 4

    def lower(self, y: np.ndarray) -> np.ndarray:
        """Lower part of the batman function."""
        val = np.zeros(y.size)

        # set s parts
        idx = np.where(np.logical_and(y <= -4, y >= -7))
        val[idx] = self._s(y[idx])
        idx = np.where(np.logical_and(y <= 7, y >= 4))
        val[idx] = self._s(y[idx])

        # set f parts
        idx = np.where(np.logical_and(y <= 4, y >= -4))
        val[idx] = self._f(y[idx])

        return val + 4

    def eval(self, y: np.ndarray) -> np.ndarray:
        """Evaluate target function.

        Parameters
        ----------
        y : np.ndarray
            Parameter values.

        Returns
        -------
        :
            Forward model evaluted in parameter values.
        """
        if y.ndim > 1:
            y = y.flatten()
        return np.array([self.upper(y), self.lower(y)]).T


class Sobol:
    """Target function for tutorial 2."""

    def __init__(self, a: Union[List, Tuple, np.ndarray]) -> None:
        """Target function for tutorial 2.

        Parameters
        ----------
        a : array_like
            Parameters for Sobol functions.
        """
        self.a = np.array(a)
        assert self.a.ndim == 1

    def eval(self, y: np.ndarray) -> np.ndarray:
        """Evaluate target function.

        Parameters
        ----------
        y : np.ndarray
            Parameter values.

        Returns
        -------
        :
            Forward model evaluted in parameter values.
        """
        if y.ndim == 1:
            y.shape = 1, -1
        assert y.ndim == 2
        assert y.shape[1] == self.a.size
        ret = (np.abs(4.0*y - 2.0) + self.a) / (1.0 + self.a)
        return np.prod(ret, axis=1).reshape(-1, 1)

    def indices(self) -> Tuple[List, np.ndarray]:
        """Compute analytical Sobol indices for target function.

        Returns
        -------
        sobol_tuples : list of tuples
            List of Sobol tuples.
        sobol_indices : np.ndarray
            Values of Sobol indices ordered the same as `sobol_tuples`.
        """
        sobol = {}
        beta = (1.0+self.a)**(-2) / 3
        var = np.prod(1.0 + beta) - 1.0
        sobol_tuples = pt.index.IndexSet(
            pt.index.tensor([1]*self.a.size)).sobol_tuples
        for sdx in sobol_tuples:
            sobol[sdx] = 1.0 / var
            for k in sdx:
                sobol[sdx] *= beta[k-1]
        sobol_indices = np.array(list(sobol.values()))
        return sobol_tuples, sobol_indices


class Sine:
    """Target function for tutorial 3."""

    def __init__(self, scales: Union[list, tuple, np.ndarray]) -> None:
        """Target function for tutorial 3.

        Parameters
        ----------
        scales : array_like
            Scales for each summand.
        """
        self.scales = np.array(scales)
        self.points = np.linspace(0, 2*np.pi, 200)
        assert self.scales.ndim == 1

    def eval(self, x: np.ndarray) -> np.ndarray:
        """Evaluate target function.

        Parameters
        ----------
        x : np.ndarray
            Parameter values.

        Returns
        -------
        :
            Forward model evaluted in parameter values.
        """
        ret = np.zeros((x.shape[0], self.points.size))
        for j, (s, xj) in enumerate(zip(self.scales, x.T)):
            ret += (s * np.reshape(xj, (-1, 1))
                    * np.sin((j+1)*self.points).reshape(1, -1))
        return ret


if __name__ == "__main__":
    pass
